package console

import (
	"context"
	"fmt"
	"io"
)

type ctxKey struct{}

// Logger is a type which may be used to write logs to a console.
type Logger struct {
	w       io.Writer
	pfx     string
	verbose bool
}

type Opt func(*Logger)

func Verbose() Opt {
	return func(l *Logger) {
		l.verbose = true
	}
}

// NewLogger constructs a Logger to write to w.
func NewLogger(w io.Writer, opts ...Opt) *Logger {
	l := &Logger{w: w}
	for _, o := range opts {
		o(l)
	}
	return l
}

// WithPrefix returns a new Logger with pfx added. Any existing prefix in l will
// be prepended.
func (l *Logger) WithPrefix(pfx string) *Logger {
	if l.pfx != "" {
		pfx = fmt.Sprintf("%s%s", l.pfx, pfx)
	}
	return &Logger{
		w:       l.w,
		pfx:     pfx,
		verbose: l.verbose,
	}
}

// Print prints a line of output to the console.
func (l *Logger) Print(output string) error {
	if _, err := l.w.Write([]byte(fmt.Sprintf("%s%s\n", l.pfx, output))); err != nil {
		return fmt.Errorf("console: could not write log: %w", err)
	}
	return nil
}

// Printf prints a line of output with formatting to the console.
func (l *Logger) Printf(format string, args ...any) error {
	return l.Print(fmt.Sprintf(format, args...))
}

// VerbosePrint prints a line of output _only if_ l was created with the Verbose
// option.
func (l *Logger) VerbosePrint(output string) error {
	if !l.verbose {
		return nil
	}
	return l.Print(output)
}

// VerbosePrintf is like VerbosePrint but for Printf.
func (l *Logger) VerbosePrintf(format string, args ...any) error {
	if !l.verbose {
		return nil
	}
	return l.Printf(format, args...)
}

// Ctx returns a child context.Context with l set as the context's logger.
func (l *Logger) Ctx(c context.Context) context.Context {
	return context.WithValue(c, ctxKey{}, l)
}

// FromCtx retrieves a Logger from c.
func FromCtx(c context.Context) *Logger {
	l := c.Value(ctxKey{})
	if l == nil {
		return NewLogger(io.Discard)
	}
	return l.(*Logger)
}
