package vegr_test

import (
	"errors"
	"fmt"
	"sync"
	"testing"
	"time"

	"git.sr.ht/~nelsam/hel/vegr"
	"git.sr.ht/~nelsam/hel/vegr/ret"
	"github.com/poy/onpar"
	"github.com/poy/onpar/expect"
)

type mockT struct {
	mu            sync.Mutex
	helperCalled  bool
	logged        bool
	logMsg        string
	fataled       bool
	failed        bool
	fatalMsg      string
	fatalPanicMsg string
	cleanuped     bool
	cleanupFn     func()
}

func (t *mockT) Helper() {
	t.mu.Lock()
	defer t.mu.Unlock()
	t.helperCalled = true
}

func (t *mockT) Failed() bool {
	t.mu.Lock()
	defer t.mu.Unlock()
	return t.failed
}

func (t *mockT) Logf(format string, args ...any) {
	t.mu.Lock()
	defer t.mu.Unlock()
	t.logged = true
	t.logMsg = fmt.Sprintf(format, args...)
}

func (t *mockT) Fatalf(format string, args ...any) {
	t.mu.Lock()
	defer t.mu.Unlock()
	t.fataled = true
	t.fatalMsg = fmt.Sprintf(format, args...)
	if t.fatalPanicMsg != "" {
		panic(t.fatalPanicMsg)
	}
}

func (t *mockT) Cleanup(f func()) {
	t.mu.Lock()
	defer t.mu.Unlock()
	t.cleanuped = true
	t.cleanupFn = f
}

func (t *mockT) didCleanup() (func(), bool) {
	t.mu.Lock()
	defer t.mu.Unlock()
	return t.cleanupFn, t.cleanuped
}

func TestPopulateReturns(t *testing.T) {
	strPtr := func() *string {
		var v string
		return &v
	}

	type testCtx struct {
		t      *testing.T
		expect expectation
		mT     *mockT
	}

	o := onpar.BeforeEach(onpar.New(t), func(t *testing.T) testCtx {
		return testCtx{t: t, expect: expect.New(t), mT: &mockT{}}
	})
	defer o.Run()

	for _, tt := range []struct {
		name       string
		prepMockT  func(*mockT)
		methodName string
		mock       func() any
		addrs      []any
		panicMsg   string
	}{
		{name: "for string types", mock: func() any { return "foo" }, panicMsg: "hel: PopulateReturns was called with incorrect mock type (string): sel: type string is neither a struct of channels nor a channel"},
		{
			name: "for non-channel fields in struct types",
			mock: func() any {
				return struct {
					Foo chan struct{}
					Bar int
				}{
					Foo: make(chan struct{}),
					Bar: 10,
				}
			},
			panicMsg: "hel: PopulateReturns was called with incorrect mock type (struct { Foo chan struct {}; Bar int }): sel: field 1 of type struct { Foo chan struct {}; Bar int } cannot be used as a mock field: sel: type int is neither a struct of channels nor a channel",
		},
		{
			name:     "for non-pointer arguments",
			mock:     func() any { return make(chan string) },
			addrs:    []any{"foo"},
			panicMsg: "hel: PopulateReturns was called with non-pointer type (string)",
		},
		{
			name:     "for nil pointer arguments",
			mock:     func() any { return make(chan string) },
			addrs:    []any{(*string)(nil)},
			panicMsg: "hel: PopulateReturns was called with nil pointer of type (*string)",
		},
		{
			name: "for closed channels",
			mock: func() any {
				ch := make(chan string)
				close(ch)
				return ch
			},
			addrs:    []any{strPtr()},
			panicMsg: "hel: PopulateReturns called on closed mock (type chan string)",
		},
		{
			name: "with a panic from t.Fatalf",
			prepMockT: func(mT *mockT) {
				mT.fatalPanicMsg = "Fail in goroutine after test has completed"
			},
			methodName: "Foo",
			mock: func() any {
				return struct {
					Foo chan string
				}{
					Foo: make(chan string, 10),
				}
			},
			addrs:    []any{strPtr()},
			panicMsg: "hel: panic calling t.Fatalf from mock method Foo: Fail in goroutine after test has completed",
		},
	} {
		tt := tt
		if tt.prepMockT == nil {
			tt.prepMockT = func(*mockT) {}
		}
		o.Group(tt.name, func() {
			o.Spec("it panics", func(test testCtx) {
				tt.prepMockT(test.mT)
				defer func() {
					r := recover()
					test.expect(r).To(not(beNil()))
					test.expect(fmt.Sprintf("%v", r)).To(equal(tt.panicMsg))
				}()
				vegr.PopulateReturns(test.mT, tt.methodName, time.Millisecond, tt.mock(), tt.addrs...)
			})

			o.Spec("it does not panic if the t has already failed", func(test testCtx) {
				tt.prepMockT(test.mT)
				defer func() {
					r := recover()
					test.expect(r).To(beNil())
				}()
				test.mT.failed = true
				vegr.PopulateReturns(test.mT, tt.methodName, time.Millisecond, tt.mock(), tt.addrs...)
			})
		})
	}

	o.Spec("it calls t.Helper() immediately", func(test testCtx) {
		defer func() {
			r := recover()
			test.expect(r).To(not(beNil()))
			test.expect(test.mT.helperCalled).To(beTrue())
		}()
		vegr.PopulateReturns(test.mT, "", time.Millisecond, "foo")
	})

	o.Spec("it fails the T if the return times out", func(test testCtx) {
		mock := make(chan string)
		var v string
		vegr.PopulateReturns(test.mT, "Foo", time.Millisecond, mock, &v)
		test.expect(test.mT.fataled).To(beTrue())
		test.expect(test.mT.fatalMsg).To(equal("hel: mock method Foo timed out after 1ms waiting for return on type (chan string)"))
	})

	o.Spec("it does not fail the T if the test has already failed", func(test testCtx) {
		mock := make(chan string)
		var v string
		done := make(chan struct{})
		go func() {
			defer close(done)
			vegr.PopulateReturns(test.mT, "Foo", 10*time.Millisecond, mock, &v)
		}()

		// This is intended to assert that if the test fails while
		// PopulateReturns is waiting on return values, PopulateReturns will not
		// call t.Fatalf (because the panic in t.Fatalf ends up hiding the
		// _real_ test failure). To ensure that this test will fail when
		// PopulateReturns only checks t.Failed _before_ waiting on the timeout
		// channel, we run PopulateReturns concurrently and update t.failed
		// after the PopulateReturns goroutine has started.
		time.Sleep(time.Millisecond)
		test.mT.mu.Lock()
		test.mT.failed = true
		test.mT.mu.Unlock()

		<-done
		test.expect(test.mT.fataled).To(not(beTrue()))
	})

	o.Spec("it sets the returned value to the address provided", func(test testCtx) {
		mock := make(chan string, 1)
		mock <- "foobar"
		var v string
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock, &v)
		test.expect(v).To(equal("foobar"))
		test.expect(test.mT.fataled).To(not(beTrue()))
	})

	o.Spec("it supports structs of channels", func(test testCtx) {
		mock := struct {
			Foo chan string
			Bar chan int
		}{
			Foo: make(chan string, 1),
			Bar: make(chan int, 1),
		}
		mock.Foo <- "foo"
		mock.Bar <- 10
		var (
			foo string
			bar int
		)
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock, &foo, &bar)
		test.expect(foo).To(equal("foo"))
		test.expect(bar).To(equal(10))
		test.expect(test.mT.fataled).To(not(beTrue()))
	})

	o.Spec("it repopulates block channels", func(test testCtx) {
		mock := struct {
			BlockChan ret.Blocker
			Foo       chan string
		}{
			BlockChan: make(ret.Blocker, 1),
			Foo:       make(chan string, 1),
		}
		mock.BlockChan <- struct{}{}
		mock.Foo <- "foo"
		var (
			foo string
		)
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock, &foo)
		test.expect(foo).To(equal("foo"))
		test.expect(test.mT.fataled).To(not(beTrue()))
		test.expect(mock.BlockChan).To(haveLen(1))
	})

	o.Spec("it does not load the same field twice", func(test testCtx) {
		mock := struct {
			Foo chan string
			Bar chan int
		}{
			Foo: make(chan string),
			Bar: make(chan int, 2),
		}
		mock.Bar <- 10
		mock.Bar <- 11
		var (
			foo string
			bar int
		)
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock, &foo, &bar)
		test.expect(bar).To(equal(10))
	})

	o.Spec("it warns about new behavior when timing out on panicker fields", func(test testCtx) {
		mock := struct {
			MockPanicker ret.Panicker
		}{
			MockPanicker: make(ret.Panicker),
		}

		defer func() {
			r := recover()
			test.expect(test.mT.logged).To(beTrue())
			test.expect(test.mT.logMsg).To(equal("hel: BREAKING CHANGE: as of v0.6, a value must also be sent on the MockPanicker field, even if a method has no return values [hint: use pers.Return, even with no return values, to ensure better forward compatibility]"))
			test.expect(r).To(beNil())
		}()
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock)
	})

	o.Spec("it does not panic on nil values from a panicker", func(test testCtx) {
		mock := struct {
			MockPanicker ret.Panicker
		}{
			MockPanicker: make(ret.Panicker, 1),
		}
		mock.MockPanicker <- nil

		defer func() {
			r := recover()
			test.expect(r).To(beNil())
		}()
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock)
	})

	o.Spec("it uses a panicker to panic the return", func(test testCtx) {
		mock := struct {
			MockPanicker ret.Panicker
		}{
			MockPanicker: make(ret.Panicker, 1),
		}
		err := errors.New("a panic")
		mock.MockPanicker <- err

		defer func() {
			r := recover()
			test.expect(r).To(equal(err))
		}()
		vegr.PopulateReturns(test.mT, "", time.Millisecond, mock)
	})

	o.Spec("it allows the test to cleanup after panicking", func(test testCtx) {
		mock := struct {
			MockPanicker ret.Panicker
		}{
			MockPanicker: make(ret.Panicker, 1),
		}
		err := errors.New("a panic")
		mock.MockPanicker <- err

		func() {
			defer func() {
				r := recover()
				test.expect(r).To(equal(err))
			}()
			vegr.PopulateReturns(test.mT, "", time.Millisecond, mock)
		}()

		cleanup, didCleanup := test.mT.didCleanup()

		test.expect(didCleanup).To(beTrue())
		done := make(chan struct{})
		go func() {
			defer close(done)
			cleanup()
		}()
		test.expect(done).To(poll(beClosed()))
	})

	o.Spec("it still reads values from other return channels when it panics", func(test testCtx) {
		mock := struct {
			MockReturnBlocker ret.Blocker
			MockPanicker      ret.Panicker
			Foo               chan string
		}{
			MockReturnBlocker: vegr.BlockChan(),
			MockPanicker:      make(ret.Panicker, 1),
			Foo:               make(chan string, 1),
		}
		err := errors.New("a panic")
		mock.MockPanicker <- err

		var foo string
		defer func() {
			r := recover()
			test.expect(r).To(equal(err))
			test.expect(foo).To(equal("foo"))
		}()
		go func() {
			// Wait a small amount of time to send the real value to ensure the
			// PopulateReturns waits for the value.
			time.Sleep(5 * time.Millisecond)
			mock.Foo <- "foo"
		}()
		vegr.PopulateReturns(test.mT, "", 10*time.Millisecond, mock, &foo)
	})

	o.Spec("it uses Cleanup to prevent tests from exiting before the mock has returned", func(test testCtx) {
		mock := make(chan string)
		var v string
		done := make(chan struct{})
		go func() {
			defer close(done)
			vegr.PopulateReturns(test.mT, "Foo", 100*time.Millisecond, mock, &v)
		}()
		for {
			select {
			case <-time.After(time.Millisecond):
				_, called := test.mT.didCleanup()
				if !called {
					continue
				}
			case <-done:
				test.t.Fatalf("expected t.Cleanup to be called")
			}
			break
		}

		fn, _ := test.mT.didCleanup()
		fn()
		select {
		case <-done:
		default:
			test.t.Fatalf("expected PopulateReturns to return when cleanup is done")
		}
	})
}
